import argparse
from typing import Any, Dict

from . import WorkerConfig, WorkerConfigDict, run

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog='worker_manager',
        description='A manager to execute a set of worker processes.'
    )
    parser.add_argument('config', help='configuration file (.py)')
    args = parser.parse_args()

    config_locals: Dict[str, Any] = {}
    with open(args.config, 'rb') as f:
        code = compile(f.read(), args.config, 'exec')
        exec(code, config_locals)  # nosec

    if 'worker_config' not in config_locals:
        raise RuntimeError('There is no local variable `worker_config` '
                           'in the configuration file')

    assert isinstance(config_locals['worker_config'], dict)

    worker_config_dict: WorkerConfigDict = dict()
    for name, config_raw in config_locals['worker_config'].items():
        worker_config_dict[name] = WorkerConfig(name=name, **config_raw)

    run(worker_config_dict)



import functools
import signal
import subprocess
import sys
import threading
import time
import traceback
from typing import Dict, List, NamedTuple, Optional

__all__ = ['run']

_POLL_INTERVAL = 0.1
_SIGTERM_TIMEOUT = 15
_MONITOR_THREAD_TIMEOUT = 5
_CONDITION_CHECK_INTERVAL = 1
_CONDITION_CHECK_TIMEOUT = 10


class WorkerConfig(NamedTuple):
    name: str
    num_workers: int
    cmd: List[str]
    start_condition_cmd: Optional[List[str]] = None


class WorkerProcessInfo(NamedTuple):
    name: str
    popen_obj: subprocess.Popen
    monitor: threading.Thread


class WorkerProcess(NamedTuple):
    worker_config: WorkerConfig
    process_info_list: List[WorkerProcessInfo]


WorkerConfigDict = Dict[str, WorkerConfig]


def run(worker_config_dict: WorkerConfigDict):
    '''Run a set of processes defined by `worker-config`

    .. note:: Example Configuration

        worker_config = {
            'manager': {
                'num_workers': 1,
                'cmd': ['python', 'launcher_manager.py']
            },
            'http-server': {
                'num_workers': 1,
                'cmd': ['gunicorn',
                        '-c', './assets/gunicorn_conf.py',
                        'launcher_http_server:server.app']
            },
            'http-server-nginx': {
                'num_workers': 1,
                'cmd': ['nginx', '-g', 'daemon off;'],
                'start_condition_cmd': ['curl',
                                        '--fail', '-s',
                                        'http://0.0.0.0:8080/ping']
            },
            'tcp-server': {
                'num_workers': os.environ.get('TCP_SERVER_NUM_WORKERS', 1),
                'cmd': ['python', 'launcher_tcp_server.py']
            }
        }
    '''

    term_signal = signal.Signals.SIGTERM
    term_event = threading.Event()

    def _signal_handler(signum, frame):
        if term_event.is_set():
            print('Forced termination of manager', flush=True)
            sys.exit(1)

        nonlocal term_signal
        term_signal = signal.Signals(signum)
        term_event.set()
        print('Signal handler called with signal :', term_signal.name,
              flush=True)

    signal.signal(signal.SIGTERM, _signal_handler)
    signal.signal(signal.SIGINT, _signal_handler)

    max_name_length = max(
        len('%s_%d' % (worker_config.name, worker_config.num_workers + 1))
        for worker_config in worker_config_dict.values()
    )

    def _p_print(prefix: str, txt: str, *, is_stdout: bool = False):
        formatstr = '%-{:d}s %s %s'.format(max_name_length)
        print(formatstr % (prefix, '|' if is_stdout else '>', txt), flush=True)

    def _process_launcher(worker_process: WorkerProcess):
        worker_config = worker_process.worker_config
        process_info_list = worker_process.process_info_list

        if worker_config.start_condition_cmd is not None:
            _p_print(
                worker_config.name,
                'Check start condition...'
            )

            while not term_event.is_set():
                try:
                    condition_result = subprocess.run(
                        worker_config.start_condition_cmd,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.STDOUT,
                        timeout=_CONDITION_CHECK_TIMEOUT
                    )
                    if condition_result.returncode == 0:
                        break
                except Exception:
                    pass

                _p_print(
                    worker_config.name,
                    'Start condition is not met'
                )

                time.sleep(_CONDITION_CHECK_INTERVAL)

        for idx in range(worker_config.num_workers):
            if term_event.is_set():
                return

            process_name = '%s_%d' % (worker_config.name, idx + 1)

            popen_obj = subprocess.Popen(
                worker_config.cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                start_new_session=True
            )

            _p_print(
                process_name,
                'Started a process with PID %d' % popen_obj.pid
            )

            monitor = threading.Thread(
                target=functools.partial(
                    _process_monitor, process_name, popen_obj
                ),
                daemon=True
            )
            monitor.start()

            process_info_list.append(
                WorkerProcessInfo(
                    name=process_name,
                    popen_obj=popen_obj,
                    monitor=monitor
                )
            )

    def _process_monitor(process_name: str, popen_obj: subprocess.Popen):
        while True:
            try:
                output = popen_obj.stdout.readline()
            except Exception:
                print('Error while reading outputs from "%s"\n%s'
                      % (process_name, traceback.format_exc()),
                      flush=True)
                break

            if not output:
                break

            _p_print(process_name, output.rstrip(), is_stdout=True)

        while True:
            retcode = popen_obj.poll()
            if retcode is not None:
                break
            time.sleep(_POLL_INTERVAL)

        _p_print(process_name, 'Terminated (retcode: %s)' % retcode)

    worker_processes: List[WorkerProcess] = []
    for worker_config in worker_config_dict.values():
        worker_process = WorkerProcess(worker_config, [])
        worker_processes.append(worker_process)

        worker_process_launcher = threading.Thread(
            target=functools.partial(_process_launcher, worker_process),
            daemon=True
        )
        worker_process_launcher.start()

    def _work_all_worker_process_info():
        for worker_process in worker_processes:
            for process_info in worker_process.process_info_list:
                yield process_info

    # Monitor the liveness of all worker processes
    done_process_info: Optional[WorkerProcessInfo] = None
    try:
        while not term_event.is_set():
            for process_info in _work_all_worker_process_info():
                if process_info.popen_obj.poll() is not None:
                    done_process_info = process_info
                    term_event.set()
                    break

                time.sleep(_POLL_INTERVAL)
        else:
            print(
                'Terminate all workers... (cuased by %s)'
                % (done_process_info.name if done_process_info else 'unknown'),
                flush=True
            )

    except Exception:
        traceback.print_exc()

    # Send stop signals to worker processes
    for process_info in _work_all_worker_process_info():
        retcode = process_info.popen_obj.poll()
        if retcode is not None:
            continue
        _p_print(process_info.name, '%s is requested' % term_signal.name)
        process_info.popen_obj.send_signal(term_signal.value)

    # Wait until all worker processes termintate
    wait_until_ts = time.time() + _SIGTERM_TIMEOUT
    while time.time() < wait_until_ts:
        for process_info in _work_all_worker_process_info():
            retcode = process_info.popen_obj.poll()
            if retcode is None:
                break
        else:
            break
        time.sleep(_POLL_INTERVAL)
    else:
        print('Timeout while waiting the termination of worker processes',
              flush=True)

    wait_until_ts = time.time() + _MONITOR_THREAD_TIMEOUT
    while time.time() < wait_until_ts:
        for process_info in _work_all_worker_process_info():
            process_info.monitor.join(0)
            if process_info.monitor.is_alive():
                break
        else:
            break
        time.sleep(_POLL_INTERVAL)
    else:
        print('Timeout while waiting the termination of monitor threads',
              flush=True)

        print('Live monitor threads:', flush=True)
        for process_info in _work_all_worker_process_info():
            if process_info.monitor.is_alive():
                print('- %s' % process_info.pname, flush=True)



import asyncio


def ensure_future_with_log(cor, *, logger, message):

    def _done_callback(fut):
        try:
            fut.result()
        except Exception:
            logger.warn(message, exc_info=True)

    asyncio.ensure_future(cor) \
        .add_done_callback(_done_callback)


async def gather_with_log(coros, *, logger):
    task_results = await asyncio.gather(*coros, return_exceptions=True)
    for result in task_results:
        try:
            if isinstance(result, Exception):
                raise result
        except Exception:
            logger.warn('Error from asyncio.gather()', exc_info=True)



import re
import sys
from os import path

from setuptools import find_packages, setup

if sys.version_info < (3, 6, 0):
    raise RuntimeError('elice-backend-components requires Python 3.6.0+')

wdir = path.abspath(path.dirname(__file__))

with open(path.join(wdir, 'ebc', '__init__.py'), encoding='utf-8') as f:
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$",
                             f.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')

with open(path.join(wdir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='elice-backend-components',

    version=version,

    description='Components for building elice-backend projects.',
    long_description=long_description,
    url='https://git.elicer.io/elice/elice-backend-components',

    author='elice.io',
    author_email='contact@elice.io',

    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X'
    ],

    packages=find_packages(),

    package_data={
        'ebc': ['py.typed']
    },

    zip_safe=False,

    install_requires=[
        'sentry-sdk>=0.7,<0.8'
    ],

    extras_require={
        'dev': [
            'bandit',
            'flake8',
            'flake8-datetimez',
            'flake8-bugbear',
            'flake8-isort',
            'mypy',
            'safety'
        ]
    }
)



import os
import sys
import traceback
from typing import Dict, Iterable, Tuple, Union

ConfigValueType = Union[None, bool, int, float, str, bytes]


def _read_config_file(config_path: str) -> Dict[str, ConfigValueType]:
    config_path_from_env = os.getenv('CONFIG_PATH')
    if config_path_from_env is not None:
        print('Replace config path with environment variable.',
              file=sys.stderr)
        config_path = config_path_from_env

    try:
        with open(config_path, encoding='utf-8') as f:
            config_source = f.read()
    except FileNotFoundError:
        print('Cannot find config file at "%s".' % config_path,
              file=sys.stderr)
        sys.exit(1)

    try:
        config_module_variables: Dict[str, ConfigValueType] = dict()
        exec(config_source, config_module_variables)  # nosec
    except Exception:
        traceback.print_exc()
        print('Failed to evaluate config file.', file=sys.stderr)
        sys.exit(1)

    config: Dict[str, ConfigValueType] = dict()
    for k, v in config_module_variables.items():
        if not k.startswith('_') and k.isupper():
            config[k] = v

    return config


class ConfigProxy:
    def __init__(self, config_dict: Dict[str, ConfigValueType]):
        self._config_dict = config_dict

    def items(self) -> Iterable[Tuple[str, ConfigValueType]]:
        return self._config_dict.items()

    def get(self, key: str, d: ConfigValueType = None) -> ConfigValueType:
        return self._config_dict.get(key, d)

    def ensure(self, key: str) -> None:
        if key not in self._config_dict:
            raise KeyError('Configuration key "%s" is missing!' % key)

    def extract(self, *keys: str):
        keys_ordered, last_idx = list(self._config_dict.keys()), -1
        for key in keys:
            self.ensure(key)
            key_idx = keys_ordered.index(key)
            if key_idx < last_idx:
                raise KeyError(
                    'The key "%s" violates the order defined in config!' % key)
            elif key_idx == last_idx:
                raise KeyError('The key "%s" is duplicated!' % key)
            else:
                last_idx = key_idx

        return ConfigProxy({key: self._config_dict[key] for key in keys})

    def __getattr__(self, key: str) -> ConfigValueType:
        return self._config_dict[key]

    def __getitem__(self, key: str) -> ConfigValueType:
        return self._config_dict[key]


def load_config(config_path: str = 'configs/default.py') -> ConfigProxy:
    return ConfigProxy(_read_config_file(config_path))
